from convert import hexToDec, binToDec, decToBinary, hexToBinary, decToHex, binToHex
import json
import sys

import os 
pathToThis = os.path.dirname(os.path.realpath(__file__))
referenceJsonPath = os.path.join(pathToThis, 'reference.json')
with open(referenceJsonPath, 'r')as f:
	jsonData = json.loads(f.read())
	Instructions = jsonData['ASM2Instr']
	wildCards = jsonData['Wildcards']
	argsDict = jsonData['ASM2Args']

counter = 0
outStr = ""
LineCounter = 1
AllFlags = {}

CustomEOF = "==="
CustomComment = '#'
CustomAddress = '$'
CustomFlag = '_'
CustomDataPrefix = '.'
CustomWildCard = '*'


class CompilerFailed(Exception):
    def __init__(self, message = 'UNDEF', line='UNDEF'):
    	print("Compiler failed" + ((': ' + message) if (message != 'UNDEF') else '') + ((' on line ' + str(line)) if (str(line) != 'UNDEF') else ''))
    	sys.exit()

# def out(binary, counter):
# 	counter += 1
# 	return binToHex(binary)

def instructionArguments(instructionLine):
	l = instructionLine
	if len(l) == 1:
		return []
	if len(l[1]) == 1: # Has a register arg
		return l[2:]
	else:
		return l[1:]

def typesOfArgs(instructionLine):
	typesList = []
	for i in instructionLine[1:]:
		if i in Instructions:
			typesList.append("Instruction")
		elif len(i) == 1 and i.isdecimal():
			typesList.append("Register")
		# elif i[:3] == '$0x':
		elif i[:1] == '$':
			typesList.append("Address")
		# elif i[:2] == '0x':
		else:
			typesList.append("Value")
		# else:
			# raise CompilerFailed("Unknown function argument type", i)
	return [None] if len(typesList) == 0 else typesList

def commandLen(instructionLine):
	l = instructionLine
	length = 0

	if len(l) == 1:
		return 1

	if len(l[1]) == 1:
		length += 1
		for number in range(2, len(l)):
			length += 1
	else:
		length += 1
		for number in range(1, len(l)):
			length += 1

	return length

def ProcessInstruction(instructionLine):
	l = instructionLine
	length = 0
	myOutStr = []
	iLoopStart = 1
	if len(l) == 1:
		return [binToHex(('0b'+Instructions[l[0]] + '00')[:10])], 1

	if len(l[1]) == 1:
		r = decToBinary(int(l[1]), length=2, pre=False)
		myOutStr.append(binToHex('0b'+Instructions[l[0]] + r))
		length += 1
		iLoopStart = 2
	else:
		myOutStr.append(binToHex(('0b'+Instructions[l[0]] + '00')[:10]))
		length += 1
	for number in range(iLoopStart, len(l)):
		if '0b' in l[number]:
			l[number] = binToHex(l[number])
		myOutStr.append(l[number])
		length += 1

	return myOutStr, length

def doPass(textToCompile, passToDo, infile, outfile, genDebugHex=False):
	# global counter, outStr, LineCounter
	global AllFlags
	counter = 0
	outStr = ""
	outList = []
	LineCounter = 1
	if passToDo == 1:
		for i in textToCompile.split("\n"):

			if i == "":
				outStr += "\n"
			else:
				l = i.split(" ")
				if l[0] == CustomEOF:
					break
				elif l[0][0] == CustomComment:
					outStr += "\n"
					pass
				elif l[0][0] == CustomAddress:
					number = hexToDec(l[0][1:])
					for num in range(int(number-counter)):
						counter += 1
					outStr += i + "\n"
				elif l[0][0] == CustomFlag:
					FlagName = l[0][1:]
					if FlagName[-1].isdecimal():
						raise CompilerFailed("Can't end a flag with a digit or decimal: " + str(FlagName), LineCounter)
					AllFlags[FlagName] = counter;
					outStr += "\n"
				elif l[0] in [CustomDataPrefix + 'db', CustomDataPrefix + 'dh', CustomDataPrefix + 'dd']:
					outStr += i + "\n"
					counter += 1
				else:
					outStr += i + "\n"
					counter += commandLen(l)
					
			LineCounter += 1

	if passToDo == 2:
		outStr = textToCompile
		lineList = textToCompile.split("\n")
		for lineNum in range(len(lineList)):
			line = lineList[lineNum]
			for element in line.split(" ")[1:]:
				found = False
				for key in AllFlags.keys():
					if key == element:
						outStr = outStr.replace(element, decToHex(AllFlags[key]), 1)
						found = True
					elif CustomAddress+key == element:
						outStr = outStr.replace(element, CustomAddress+decToHex(AllFlags[key]), 1)
						found = True
					elif key in element:
						valueD = AllFlags[key]
						# print(element, element.replace(key, ''), element.replace(key, '').replace(CustomAddress, ''))
						numberD = int(element.replace(key, '').replace(CustomAddress, ''))
						outStr = outStr.replace(element, (CustomAddress if element[0] == CustomAddress else '') + decToHex(valueD + numberD), 1)
						found = True
				if not ('0x' in element or '0b' in element) and not element.isdecimal():
					if not found:
						flagToError = element.replace(CustomAddress, '')
						for i in range(len(list(flagToError))):
							thisChar = flagToError[len(list(flagToError))-i-1]
							if not thisChar.isdecimal():
								flagToError = flagToError[:len(list(flagToError))-i]
								break
						raise CompilerFailed(f"Undeclared flag: '{flagToError}'", lineNum+1)

	if passToDo == 3:
		lineList = textToCompile.split("\n")
		for lineNum in range(len(lineList)):
			line = lineList[lineNum]
			if line == '':
				continue
			elements = line.split(" ")
			# if line[0] == '*':
			# if len(elements[0]) == 2:
			if (CustomWildCard + elements[0].replace(CustomWildCard, '')) in wildCards:
				iArgs = instructionArguments(elements)
				# print(elements, iArgs)
				try:
					if iArgs[0][0] == CustomAddress: # First letter of first arg is address marker
						decided = wildCards[CustomWildCard + elements[0].replace(CustomWildCard, '')]["Address"]
					else: # First letter of first arg is address marker
						try:
							decided = wildCards[CustomWildCard + elements[0].replace(CustomWildCard, '')]["Value"]
						except:
							raise CompilerFailed(f"Unrecognized instruction: {elements[0]}", lineNum+1)
				except KeyError:
					iArgsTypes = [argument for argument in typesOfArgs(elements) if argument != 'Register']
					Targets = [('None' if target == '' else target) for target in wildCards[CustomWildCard + elements[0].replace(CustomWildCard, '')]]
					raise CompilerFailed("The wildcard instruction " + CustomWildCard + elements[0].replace(CustomWildCard, '') + " requires on of argument sets [" + (', '.join(Targets)) + "], not [" + (', '.join(iArgsTypes)) + "]", lineNum+1)
				except IndexError:
					raise CompilerFailed("Wildcard instructions require at least one argument other than register", lineNum+1)

				# print(decided)
				lineList[lineNum] = decided + ' ' + ' '.join(lineList[lineNum].split(' ')[1:])
		outStr = '\n'.join(lineList)

	if passToDo == 4:
		lineList = textToCompile.split("\n")
		for lineNum in range(len(lineList)):
			line = lineList[lineNum]
			if line == '':
				continue
			elements = line.split(" ")
			if elements[0] in Instructions:
				iArgsTypes = typesOfArgs(elements)
				if not iArgsTypes == argsDict[elements[0]]: # If incorrect arguments
					if argsDict[elements[0]] == [None]:
						raise CompilerFailed("The instruction " + elements[0] + " requires no arguments", lineNum+1)
					else:
						raise CompilerFailed("The instruction " + elements[0] + " requires arguments '" + (', '.join(argsDict[elements[0]])) + "', not '" + (', '.join(iArgsTypes)) + "'", lineNum+1)
				lineList[lineNum] = ' '.join([element.replace('$', '') for element in elements])

		outStr = '\n'.join(lineList)

	if passToDo == 5:
		lineList = textToCompile.split("\n")
		for lineNum in range(len(lineList)):
			line = lineList[lineNum]
			if line != "":
				elements = line.split(" ")
				if elements[0][0] == CustomAddress:
					number = hexToDec(elements[0][1:])
					for num in range(int(number-counter)):
						outList.append(binToHex(('0b'+Instructions["NOP"] + '11')[:10]))
						counter += 1
				elif elements[0] == CustomDataPrefix + 'db':
					counter += 1
					outList.append(binToHex(elements[1]))
				elif elements[0] == CustomDataPrefix + 'dh':
					counter += 1
					outList.append(elements[1])
				elif elements[0] == CustomDataPrefix + 'dd':
					counter += 1
					outList.append(decToHex(elements[1]))
				else:
					if not elements[0] in Instructions:
						raise CompilerFailed(f"Unkown Instruction {elements[0]}", lineNum)
					hexes, length = ProcessInstruction(elements)
					counter += length
					outList += hexes
			LineCounter += 1
		outStr = ''.join([outList[i].lower() + ('\n' if i % 8 == 7 else ' ') for i in range(len(outList))])

	if genDebugHex:
		nameOfFile = "Debug." + str(passToDo) + ('.hex' if passToDo == 5 else '.asm')
		with open(os.path.join(pathToThis, nameOfFile), "w+") as f:
			f.write(outStr)
	if outList != []:
		return outStr, outList
	return outStr

def genSyntax(genSyntaxPath):
	with open(genSyntaxPath, 'r') as f:
		genSyntaxTemplate = f.read()

	genSyntaxData = genSyntaxTemplate
	genSyntaxData = genSyntaxData.replace(r"%WILDCARDS%", '|'.join(list(wildCards.keys())).replace('*', ''))
	genSyntaxData = genSyntaxData.replace(r"%COMMANDS%", '|'.join(list(Instructions.keys())))
	genSyntaxData = genSyntaxData.replace(r"%EOF%", CustomEOF)
	genSyntaxData = genSyntaxData.replace(r"%Comment%", CustomComment)
	genSyntaxData = genSyntaxData.replace(r"%Address%", CustomAddress)
	genSyntaxData = genSyntaxData.replace(r"%Flag%", CustomFlag)
	genSyntaxData = genSyntaxData.replace(r"%DataPrefix%", CustomDataPrefix)
	genSyntaxData = genSyntaxData.replace(r"%WildCard%", CustomWildCard)

	with open(genSyntaxPath.replace('.template', ''), 'w+') as f:
		f.write(genSyntaxData)

def main(silent=False, genSyntaxPath="C:\\Users\\jimmy\\AppData\\Roaming\\Sublime Text 3\\Packages\\User\\MyASM\\MyASM.sublime-syntax.template", genDebugHex=False, infile="prog.asm", outfile="UNSET"):

	# global counter, outStr, LineCounter
	global AllFlags

	if infile[-4:] != '.asm':
		print(f"Warning: input file ends with '{infile[-4:]}' instead of '.asm' - output file will be '{infile[:-4]}.hex'")
		outfile = infile + '.hex'
	else:
		outfile = infile[:-4] + '.hex'


	if not silent:
		print("Compiling '" + infile + "' to '" + outfile + "'")

	if genSyntaxPath != False:
		print("Generating Syntax File")
		genSyntax(genSyntaxPath)

	AllFlags = {}
	with open(infile, "r") as f:
		fileIn = f.read().replace("  ", " ")

	pass1 = doPass(fileIn, 1, infile, outfile, genDebugHex=genDebugHex)
	pass2 = doPass(pass1, 2, infile, outfile, genDebugHex=genDebugHex)
	pass3 = doPass(pass2, 3, infile, outfile, genDebugHex=genDebugHex)
	pass4 = doPass(pass3, 4, infile, outfile, genDebugHex=genDebugHex)
	Final, FinalList = doPass(pass4, 5, infile, outfile, genDebugHex=genDebugHex)

	with open(outfile, "w+") as f:
		f.write(Final)

	if not silent:
		if len(AllFlags.keys()) > 0:
			print('Flags:')
			for key in AllFlags.keys():
				print(' -', key, decToHex(AllFlags[key]), '/ 0d' + str(AllFlags[key]))
		print(f"Size: {len(FinalList)}b")
		print("Done!")

if __name__ == '__main__':
	args = sys.argv[1:]

	if len(args) == 1:
		infile = args[0]
		main(infile = infile)