import logging
import copy
import time
import json
logging.basicConfig(level=logging.WARN)
import notional
import os 
pathToThis = os.path.dirname(os.path.realpath(__file__))
referenceJsonPath = os.path.join(pathToThis, 'reference.json')

Settings = {
	"auth":"secret_YF8BHJ7rqLhhAHoFJngXkhYvjIXVIzdt7Veti0VlAPh",
	"DATA_COLUMNS": [
		"ID",
		"Instruction",
		"Name",
		"ASM",
		"Args",
		"Test",
		"Micro ASM",
		"Micro Code",
	],
	'IDs':{
		"TABLE":			"20c83f73-4fa6-4ef4-98d4-c15182ddf920",
		"MSBMicroCode":		"27e54bf004604579bbb25db7fba603f4",
		"LSBMicroCode":		"0a5e3a76ddfc430fbc83457768cd3bf5",
		"Shorthand":		"c6617751086a4fcc99de38fc9a696f95",
		"MiniCommands":		"dc506114758a4ac3b1f1fe5c4b1affa2",
		"MathArguments":	"396a45a2afab4fdfb0cb1db8e554a8c9",
	}
}

# from notional import blocks, types
notion = notional.connect(auth =Settings['auth'])


MicroCodeDictionary = {}

ShortHandDictionary = {}

def HighlightRed(rowList, thingToHighlight, rowObj, column = Settings['DATA_COLUMNS'].index("Micro ASM")):
	global symbol
	symbol = 'H'
	#print("    HighLight - ", end='')
	CorrectObj = "FAILED"
	CorrectObjI = -1
	for i in range(len(rowList[column])):
		obj = rowList[column][i]
		if thingToHighlight in obj.text.content:
			CorrectObj = obj
			CorrectObjI = i
			break
	if CorrectObj != "FAILED":
		pre, post = str(CorrectObj).split(thingToHighlight, 1)
		#print(pre + "|" + thingToHighlight + "|" + post)

		MidObj = copy.deepcopy(CorrectObj)
		MidObj.text.content = thingToHighlight
		MidObj.annotations.color = 'red_background'
		MidObj.annotations.underline = True
		rowList[column][CorrectObjI] = MidObj

		if pre != '':
			PreObj = copy.deepcopy(CorrectObj)
			PreObj.text.content = pre
			PreObj.annotations.color = 'default'
			PreObj.annotations.underline = False
			rowList[column].insert(CorrectObjI, PreObj)

		if post != '':
			PostObj = copy.deepcopy(CorrectObj)
			PostObj.text.content = post
			PostObj.annotations.color = 'default'
			PostObj.annotations.underline = False
			rowList[column].insert(CorrectObjI+(2 if pre != '' else 1), PostObj)
		rowObj.table_row.cells = rowList
		notion.blocks.update(rowObj)
	else:
		#print("Failed to separate")
		pass

def processCommand(command, rowList, rowObj):
	global MicroCodeDictionary, ShortHandDictionary
	# print(json.dumps(MicroCodeDictionary, indent=4))
	# print(json.dumps(ShortHandDictionary, indent=4, sort_keys=True))
	if command in ShortHandDictionary:
		if ShortHandDictionary[command][:2] == "MC":
			return MicroCodeDictionary["MSB"]["Mini Command"] + ShortHandDictionary[command][2:]
		return ShortHandDictionary[command]
	elif command[:2] == "MC":
		return MicroCodeDictionary["MSB"]["Mini Command"] + command[2:]
	elif command[:2] == "MF":
		return MicroCodeDictionary["MSB"]["Math"] + command[2:]
	elif "→" in command:
		MSBKey, LSBKey = command.split('→')
		MSB, LSB = "", ""
		shouldHighlight = False
		try:
			if type(ShortHandDictionary[MSBKey]) == str:
				MSB = ShortHandDictionary[MSBKey]
			elif type(ShortHandDictionary[MSBKey]) == list:
				MSB = ShortHandDictionary[MSBKey][0]
			# print("First part is", MSB)
			MSB = MicroCodeDictionary["MSB"][MSB]
			# print("(MC)", MSB)
		except KeyError:
			MSB = "XXXX"
			shouldHighlight = True

		try:
			if type(ShortHandDictionary[LSBKey]) == str:
				LSB = ShortHandDictionary[LSBKey]
			elif type(ShortHandDictionary[LSBKey]) == list:
				LSB = ShortHandDictionary[LSBKey][1]
			# print("Second part is", LSB)
			LSB = MicroCodeDictionary["LSB"][LSB]
			# print("(MC)", LSB)
		except KeyError:
			LSB = "XXXX"
			shouldHighlight = True

		if shouldHighlight:
			HighlightRed(rowList, command, rowObj)
		return MSB + LSB
	else:
		print(f"Inoperable command {command}")


def updateRow(rowToUpdate, rowObj):
	global symbol
	beforeCode = copy.deepcopy(rowToUpdate[Settings['DATA_COLUMNS'].index("Micro Code")][0].text.content)
	OutputList = []
	MCIs = ""
	for textSection in rowToUpdate[Settings['DATA_COLUMNS'].index("Micro ASM")]:
		textSection.annotations.underline = False
		textSection.annotations.color = 'default'
		MCIs += textSection.text.content
	MCIs = MCIs.split(",")
	for command in MCIs:
		CmdOut = processCommand(command, rowToUpdate, rowObj)
		OutputList.append(CmdOut)
	try:
		rowToUpdate[Settings['DATA_COLUMNS'].index("Micro Code")][0].text.content = " ".join(OutputList)
		if rowToUpdate[Settings['DATA_COLUMNS'].index("Micro Code")][0].text.content != beforeCode:
			# print("update")
			rowObj.table_row.cells = rowToUpdate
			notion.blocks.update(rowObj)
		else:
			symbol = '='
	except TypeError:
		symbol = 'B'


def combineTextList(textList):
	combinedText = ""
	if len(textList) == 0:
		raise IndexError
	for textSection in textList:
		combinedText += textSection.text.content
	return combinedText

def chkNone(stringy):
	if stringy == 'None':
		return None
	else:
		return stringy

# def testIndex():
	
def index(Printer=True):
	if Printer:
		print("Indexing...", end='', flush=True)
	TABLE = notion.blocks.retrieve(Settings['IDs']['TABLE'])
	tableChildren = notion.blocks.children.list(TABLE)
	
	ASM2Instr = {}
	Instr2MC = {}
	ASM2Args = {}
	WildCards = {}
	
	i = 0
	for row in tableChildren:
		rowChildren = row.table_row.cells
		if (i != 0) and (len(rowChildren[Settings['DATA_COLUMNS'].index("Instruction")]) != 0) and (len(rowChildren[Settings['DATA_COLUMNS'].index("Micro Code")]) != 0):
			Instr = combineTextList(rowChildren[Settings['DATA_COLUMNS'].index("Instruction")])
			ASM   = combineTextList(rowChildren[Settings['DATA_COLUMNS'].index("ASM")])
			Args  = combineTextList(rowChildren[Settings['DATA_COLUMNS'].index("Args")])
			MC    = combineTextList(rowChildren[Settings['DATA_COLUMNS'].index("Micro Code")])
			
			if not ASM in ASM2Instr.keys():
				ASM2Instr[ASM] = Instr
			else:
				print("Duplicate Assembly Command")
				HighlightRed(rowChildren, ASM, row, column = Settings['DATA_COLUMNS'].index("ASM"))

			argsList = Args.split(' ')
			if argsList[0][0] == '*':
				argumentToGet = 2 if argsList[1] == 'Register' else 1
				try:
					WildCards[argsList[0]][argsList[argumentToGet]] = ASM
				except KeyError:
					WildCards[argsList[0]] = {}
					try:
						WildCards[argsList[0]][argsList[argumentToGet]] = ASM
					except IndexError:
						WildCards[argsList[0]][''] = ASM
			ASM2Args[ASM] = [chkNone(argument) for argument in argsList if argument[0] != '*']
				
			if not Instr in Instr2MC.keys():
				Instr2MC[Instr] = MC
			else:
				print("Duplicate Instruction Code")
				HighlightRed(rowChildren, Instr, row, column = Settings['DATA_COLUMNS'].index("Instruction"))

		i+=1
		
	MSBBlock = notion.blocks.retrieve(Settings['IDs']['MSBMicroCode'])
	MSBList = combineTextList(MSBBlock.paragraph.rich_text).split("\n")
	MSBDict = {}
	for MSBEntry in MSBList:
		if not ':' in MSBEntry:
			continue
		MSBEntryList = MSBEntry.split(':')
		if MSBEntryList[1].strip() != "":
			MSBDict[MSBEntryList[1].strip()] = MSBEntryList[0].strip()
	# print(json.dumps(MSBDict, indent=4))
	LSBBlock = notion.blocks.retrieve(Settings['IDs']['LSBMicroCode'])
	LSBList = combineTextList(LSBBlock.paragraph.rich_text).split("\n")
	LSBDict = {}
	for LSBEntry in LSBList:
		if not ':' in LSBEntry:
			continue
		LSBEntryList = LSBEntry.split(':')
		if LSBEntryList[1].strip() != "":
			LSBDict[LSBEntryList[1].strip()] = LSBEntryList[0].strip()
	# print(json.dumps(LSBDict, indent=4))
	
	SHBlock = notion.blocks.retrieve(Settings['IDs']['Shorthand'])
	SHList = combineTextList(SHBlock.paragraph.rich_text).split("\n")
	SHDict = {}
	for SHEntry in SHList:
		if not '=' in SHEntry:
			continue
		SHEntryList = SHEntry.split('=')
		SHValue = SHEntryList[1].strip()
		if '/' in SHValue:
			SHValue = SHValue.split('/')
			SHValue[0], SHValue[1] = chkNone(SHValue[0].strip()), chkNone(SHValue[1].strip())
		else:
			SHValue = chkNone(SHValue.strip())
		SHDict[SHEntryList[0].strip()] = SHValue
	# print(json.dumps(SHDict, indent=4, sort_keys=True))
	
	MCBlock = notion.blocks.retrieve(Settings['IDs']['MiniCommands'])
	MCList = combineTextList(MCBlock.paragraph.rich_text).split("\n")
	MCDict = {}
	for MCEntry in MCList:
		if not ':' in MCEntry:
			continue
		MCEntryList = MCEntry.split(':')
		if MCEntryList[1].strip() != "":
			MCDict[MCEntryList[0].strip()] = MCEntryList[1].strip()
	
	MathBlock = notion.blocks.retrieve(Settings['IDs']['MathArguments'])
	MathList = combineTextList(MathBlock.paragraph.rich_text).split("\n")
	MathDict = {}
	for MathEntry in MathList:
		if not ':' in MathEntry:
			continue
		MathEntryList = MathEntry.split(':')
		if MathEntryList[1].strip() != "":
			MathDict[MathEntryList[0].strip()] = MathEntryList[1].strip().replace("←", "L").replace("→", "R")
	
	with open(referenceJsonPath, "w+") as f:
		f.write(json.dumps({"ASM2Instr":ASM2Instr, "Instr2MC":Instr2MC, "MCDict":{"MSB":MSBDict, "LSB":LSBDict}, "SHDict":SHDict, "Wildcards": WildCards, "ASM2Args": ASM2Args, "MiCDict":MCDict, "MathDict":MathDict}, indent=4))
	if Printer:
		print("Done.")

try:
	with open(referenceJsonPath, 'r') as f:
		jsonDict = json.loads(f.read())
		MicroCodeDictionary = jsonDict['MCDict']
		ShortHandDictionary = jsonDict['SHDict']
except:
	print("Failed to load reference, re-indexing...", end='')
	index(Printer=False)
	print("Done.")

cache = {}
symbol = '#'

def main(Printer=True, renumber=False): # renumber should be True, False, or 'Only'
	global symbol
	i=0
	TABLE = notion.blocks.retrieve(Settings['IDs']['TABLE'])
	tableChildren = notion.blocks.children.list(TABLE)
	# os.system('cls')
	if Printer:
		if renumber == 'Only':
			print("Renumbering!")
		else:
			print("Refactoring and Compiling..." + ("(And renumbering)" if renumber == True else ''))
	for row in tableChildren:
		symbol = '#'
		if renumber != 'Only':
			try:
				if ((i in RowsToCheck and UpdateRows) or CheckAllRows) and str(row.table_row.cells[1][0])[0] in ['0', '1']:
					if i in cache.keys():
						if row.table_row.cells != cache[i]:
							cache[i] = row.table_row.cells
							updateRow(row.table_row.cells, row)
						else:
							symbol = 'C'
					else:
						cache[i] = row.table_row.cells
						updateRow(row.table_row.cells, row)
				else:
					symbol = '-'
			except IndexError:
				symbol = '-'


		# if i in [22]:
		# 	symbol = '*'

		if renumber == True or renumber == 'Only':
			try:
				row.table_row.cells[Settings['DATA_COLUMNS'].index("ID")][0].text.content = (str(i) if i != 0 else "ID")
				notion.blocks.update(row)
			except IndexError:
				symbol = 'B'
		if Printer:
			print(symbol, end='\n' if i%8==7 else '', flush=True)
		i+=1
	# print(cache)
	if Printer:
		print()
		print("Done!")
		
def mainLoop():
	mainLoopi = 0
	while Loop or mainLoopi == 0:
		main()
		if mainLoopi%10 == 0:
			index()
		time.sleep(1)
		mainLoopi += 1


RowsToCheck = []
UpdateRows = True
CheckAllRows = True

Loop = False


if __name__ == '__main__':
	# mainLoop()
	# main(renumber=True)
	index()