import componentTypes
from convert import preToDec, hexToDec, binToDec, decToBin, hexToBin, decToHex, binToHex
import os
import sys
import json
import math
try:
	import curses
	HAVE_CURSES = True
except:
	HAVE_CURSES = False
   
if HAVE_CURSES:
	stdscr = curses.initscr()
	# curses.nocbreak()
	# stdscr.keypad(False)
	# curses.echo()

args = sys.argv[1:]

program = args[-1] #or '../Tests/Math.hex'

with open("../reference.json", 'r') as f:
	jsonData = json.loads(f.read())
	MicroCodeDefinitionsR = jsonData["MCDict"]
	MiniCommandsDictR = jsonData["MiCDict"]
	MiniCommandsDict = {}
	MathCommandsDictR = jsonData["MathDict"]
	MathCommandsDict = {}
	commandToAsmNameR = jsonData["ASM2Instr"]
	commandToAsmName = {}
	MicroCodeDefinitions = {"MSB":{}, "LSB":{}}

for key, value in MicroCodeDefinitionsR["MSB"].items():
	MicroCodeDefinitions["MSB"][binToHex(value, pre=False, length=1)] = key
for key, value in MicroCodeDefinitionsR["LSB"].items():
	MicroCodeDefinitions["LSB"][binToHex(value, pre=False, length=1)] = key
for key, value in MiniCommandsDictR.items():
	MiniCommandsDict[binToHex(key, pre=False, length=1)] = value
for key, value in MathCommandsDictR.items():
	MathCommandsDict[binToHex(key, pre=False, length=1)] = value
for key, value in commandToAsmNameR.items():
	commandToAsmName[value] = key
# print(json.dumps(MathCommandsDict, indent=2))
# raise ValueError
with open("../ROM.hex", 'r') as f:
	MicroCodeROM = f.read()
with open(program, 'r') as f:
	program = f.read()

RAM						= componentTypes.RAM(program.split())
Micro_ROM				= componentTypes.RAM(MicroCodeROM.split("\n"))
Main_Counter 			= componentTypes.counter((2**2)-1)
Micro_Counter 			= componentTypes.counter((2**4)-1)
PC 						= componentTypes.IncByte(255)
MAR						= componentTypes.Byte()
MDR						= componentTypes.Byte()
MWR						= componentTypes.Byte()
Register0				= componentTypes.Byte()
Register1				= componentTypes.Byte()
Register2				= componentTypes.Byte()
Register3				= componentTypes.Byte()
RegisterA				= componentTypes.Byte()
RegisterB				= componentTypes.Byte()
RegisterY				= componentTypes.Byte()
RegisterFlags			= componentTypes.Flags()
InstR					= componentTypes.Byte()
AddrR					= componentTypes.Byte()
ValuR					= componentTypes.Byte()

loops = 1
clockPulses = 0
InstructionsExecuted = 0
prefix = ':'

# ┌─┬─┐
# │ │ │
# ├─┼─┤
# │ │ │
# └─┴─┘
UI = """
┌──────────────┬──────────────────────────┐
│ Registers    │ Counters                 │
│ ──────────── │ ──────────────────────── │
│ R0      R0// │ Program Counter     PC// │
│ R1      R1// │ Micro Counter       IC// │
│ R2      R2// │ Main Counter        MC// │
│ R3      R3// │     MainCounterAction/// │
│ ──────────── │                          │
│ RA      RA// │                          │
│ RB      RB// ├──────────────────────────┤
│ RY      RY// │ Special Registers        │
├──────────────┤ ──────────────────────── │
│ Mem Rs       │ Instruction  IR///////// │
│ ──────────── │ Address             AR// │
│ MDR     MDR/ │ Value               VR// │
│ MAR     MAR/ │ Flags:          XXJROCHN │
│ MWR     MWR/ │                 RF////// │
├──────────────┴──────────────────────────┤
│ Current Instruction ASM/                │
│ A///-////-////-////-////-////-////-//// │
│ B----^----^----^----^----^----^----^--- │
│ C///-////-////-////-////-////-////-//// │
│ D----^----^----^----^----^----^----^--- │
│ ─────────────────────────────────────── │
│ Micro Code Translation                  │
│ F////////////////////////////////////// │
│ T////////////////////////////////////// │
├─────────────────────────────────────────┤
│ Clock Cycles:       CLK///////////////  │
│ Instructions:       INST//////////////  │
└─────────────────────────────────────────┘
"""

class EmulatorError(Exception):
    def __init__(self, message = 'UNDEF'):
    	print("Emulator failed [Clock cycle " + str(clockPulses) + ']' + ((': ' + message) if (message != 'UNDEF') else ''))
    	sys.exit()

def Twos(num):
	return binToDec(''.join(['1' if i == '0' else '0' for i in decToBin(num-1, pre=False)]))
def Invert(num):
	return binToDec(''.join(['1' if i == '0' else '0' for i in decToBin(num, pre=False)]))

def formatLen(thing, length=4, align="Right"):
	if align == "Left":
		return ((str(thing) + " "*length)[:length])
	else:
		return ((" "*length + str(thing))[-length:])

def splitInstruciton(inst):
	return binToDec(decToBin(inst, pre=False)[:6]), binToDec(decToBin(InstR.Read(), pre=False)[6:])

def checkForErrors():
	try:
		commandToAsmName[decToBin(InstR.Read(), pre=False)[:6]]
	except KeyError:
		EmulatorError(f"Could not find binary instruction {decToBin(InstR.Read(), pre=False)[:6]} in references dictionary. Try re-indexing, or your ROM/Docs may be improperly formatted")

def PrintDiagnostic():
	UITemp = str(UI)
	instLine = Micro_ROM.Read(splitInstruciton(InstR.Read())[0])
	MicroInstruction = Micro_ROM.Read(splitInstruciton(InstR.Read())[0]).split(" ")[Micro_Counter.Read()]
	registerArgument = splitInstruciton(InstR.Read())[1]

	if Main_Counter.Read() == 0:
		MainCounterAction = "Increment PC"
	if Main_Counter.Read() == 1:
		MainCounterAction = "Read at PC"
	if Main_Counter.Read() == 2:
		MainCounterAction = "Load Instruction"
	if Main_Counter.Read() == 3:
		MainCounterAction = "Execute"
		MSBThing = MicroCodeDefinitions["MSB"][MicroInstruction[2]]
		if MSBThing == "Mini Command":
			MSBDebug = formatLen("Mini Command: ", 39, align="Left")
			LSBDebug = formatLen("  " + MiniCommandsDict[MicroInstruction[3]], 39, align="Left")
		if MSBThing == "Math":
			MSBDebug = formatLen("Math Operation: ", 39, align="Left")
			LSBDebug = formatLen("  " + MathCommandsDict[MicroInstruction[3]] + " R" + str(registerArgument), 39, align="Left")
		else:
			if MSBThing not in ["Mini Command", "Math"]:
				LSBThing = MicroCodeDefinitions["LSB"][MicroInstruction[3]]
				if MSBThing == "Register":
					MSBDebug = "From " + formatLen(MSBThing + ' ' + str(splitInstruciton(InstR.Read())[1]), 34, align="Left")
				else:
					MSBDebug = "From " + formatLen(MSBThing, 34, align="Left")
				if LSBThing == "Register":
					LSBDebug = "To   " + formatLen(LSBThing + ' ' + str(splitInstruciton(InstR.Read())[1]), 34, align="Left")
				else:
					LSBDebug = "To   " + formatLen(LSBThing, 34, align="Left")
	else:
		MSBDebug = "None" + (" "*35)
		LSBDebug = " "*39
	checkForErrors()
	replDict = {
		"MC//": formatLen(decToHex(Main_Counter.Read()), 4),
		"MainCounterAction///":formatLen(MainCounterAction, 20, align="Left"),
		"IC//": formatLen(decToHex(Micro_Counter.Read()), 4),
		"PC//": formatLen(decToHex(PC.Read()), 4),
		"R0//": formatLen(decToHex(Register0.Read()), 4),
		"R1//": formatLen(decToHex(Register1.Read()), 4),
		"R2//": formatLen(decToHex(Register2.Read()), 4),
		"R3//": formatLen(decToHex(Register3.Read()), 4),
		"RA//": formatLen(decToHex(RegisterA.Read()), 4),
		"RB//": formatLen(decToHex(RegisterB.Read()), 4),
		"RY//": formatLen(decToHex(RegisterY.Read()), 4),
		"RF//////": formatLen(decToBin(RegisterFlags.Read(), pre=False), 8),
		"MAR/": formatLen(decToHex(MAR.Read()), 4),
		"MDR/": formatLen(decToHex(MDR.Read()), 4),
		"MWR/": formatLen(decToHex(MWR.Read()), 4),
		"IR/////////": decToBin(InstR.Read())[:8] + "|" + decToBin(InstR.Read(), pre=False)[6:],
		"AR//": formatLen(decToHex(AddrR.Read()), 4),
		"VR//": formatLen(decToHex(ValuR.Read()), 4),
		"ASM/": formatLen(commandToAsmName[decToBin(InstR.Read(), pre=False)[:6]], 4, align="Left") if Main_Counter.Read() == 3 else " "*4,
		"A///-////-////-////-////-////-////-////": instLine[:39]   if Main_Counter.Read() == 3 else "None" + (" "*35),
		"B----^----^----^----^----^----^----^---": (" "*(Micro_Counter.Read()*5) + "^^^^" + (" "*(35-(Micro_Counter.Read()*5)))) if Main_Counter.Read() == 3 and Micro_Counter.Read()<8 else " "*39,
		"C///-////-////-////-////-////-////-////": instLine[40:79] if Main_Counter.Read() == 3 else " "*39,
		"D----^----^----^----^----^----^----^---": (" "*((Micro_Counter.Read()*5)-40) + "^^^^" + (" "*(35-((Micro_Counter.Read()*5)-40)))) if Main_Counter.Read() == 3 and Micro_Counter.Read()>=8 else " "*39,
		"F//////////////////////////////////////": MSBDebug,
		"T//////////////////////////////////////": LSBDebug,
		"CLK///////////////": formatLen(clockPulses, 18),
		"INST//////////////": formatLen(InstructionsExecuted, 18),
	}
	for key, value in replDict.items():
		UITemp = UITemp.replace(key, value)
	if not HAVE_CURSES:
		os.system("clear")
		print(UITemp)
	else:
		stdscr.addstr(0, 0, UITemp)
		stdscr.refresh()

def SetJump(thing):
	if thing:
		RegisterY.Write(1)
		RegisterFlags.setFlag("J", True)
	else:
		RegisterY.Write(0)
		RegisterFlags.setFlag("J", False)

def clock():
	global clockPulses, InstructionsExecuted
	clockPulses += 1
	if Main_Counter.Read() == 0:
		InstructionsExecuted += 1
		PC.Inc()
		Micro_Counter.Reset()
		Main_Counter.Inc()
		return
	if Main_Counter.Read() == 1:
		MAR.Write(PC.Read())
		MDR.Write(preToDec(RAM.Read(MAR.Read())))
		Main_Counter.Inc()
		return
	if Main_Counter.Read() == 2:
		InstR.Write(MDR.Read())
		Main_Counter.Inc()
		return
	if Main_Counter.Read() == 3:
		MicroInstruction = Micro_ROM.Read(splitInstruciton(InstR.Read())[0]).split(" ")[Micro_Counter.Read()]
		stdscr.addstr(curses.LINES-2,0, "                                       ")
		instrP1 = MicroCodeDefinitions["MSB"][MicroInstruction[2]]
		registerArgument = splitInstruciton(InstR.Read())[1]
		if instrP1 == "Mini Command":
			MiC = MiniCommandsDict[MicroInstruction[3]]
			if MiC == "Inc PC":
				PC.Inc()
			elif MiC == "Check & Jump":
				if RegisterFlags.getFlag("J"):
					PC.Write(AddrR.Read()-1)
			elif MiC == "Halt":
				RegisterFlags.setFlag("H", True)
			elif MiC == "DONE":
				Main_Counter.Reset()
			else:
				raise ValueError("Unknown Mini Command " + MiniCommandsDict[MicroInstruction[3]])
		elif instrP1 == "Math":
			if MathCommandsDict[MicroInstruction[3]] == "ADD":
				RegisterY.Write(RegisterA.Read() + RegisterB.Read())
			elif MathCommandsDict[MicroInstruction[3]] == "SUB":
				RegisterY.Write(RegisterA.Read() - RegisterB.Read())
			elif MathCommandsDict[MicroInstruction[3]] == "INC":
				RegisterY.Write(RegisterA.Read() + 1)
			elif MathCommandsDict[MicroInstruction[3]] == "DEC":
				RegisterY.Write(RegisterA.Read() - 1)
			elif MathCommandsDict[MicroInstruction[3]] == "AND":
				RegisterY.Write(RegisterA.Read() & RegisterB.Read())
			elif MathCommandsDict[MicroInstruction[3]] == "NAND":
				RegisterY.Write(~(RegisterA.Read() & RegisterB.Read()))
			elif MathCommandsDict[MicroInstruction[3]] == "OR":
				RegisterY.Write(RegisterA.Read() | RegisterB.Read())
			elif MathCommandsDict[MicroInstruction[3]] == "XOR":
				RegisterY.Write(RegisterA.Read() ^ RegisterB.Read())
			elif MathCommandsDict[MicroInstruction[3]] == "SHIFTL":
				RegisterY.Write(RegisterA.Read() << 1)
			elif MathCommandsDict[MicroInstruction[3]] == "SHIFTR":
				RegisterY.Write(RegisterA.Read() >> 1)
			elif MathCommandsDict[MicroInstruction[3]] == "NEG (2s)":
				RegisterY.Write(Twos(RegisterA.Read()))
			elif MathCommandsDict[MicroInstruction[3]] == "NOT":
				RegisterY.Write(Invert(RegisterA.Read()))
			elif MathCommandsDict[MicroInstruction[3]] == "==0":
				SetJump(RegisterA.Read() == 0)
			elif MathCommandsDict[MicroInstruction[3]] == "==":
				SetJump(RegisterA.Read() == RegisterB.Read())
			elif MathCommandsDict[MicroInstruction[3]] == "A<B":
				SetJump(RegisterA.Read() < RegisterB.Read())
			elif MathCommandsDict[MicroInstruction[3]] == "A>B":
				SetJump(RegisterA.Read() > RegisterB.Read())
			else:
				raise ValueError("Unknown Math Command " + MathCommandsDict[MicroInstruction[3]])
		else:
			instrP2 = MicroCodeDefinitions["LSB"][MicroInstruction[3]]
			TEMP_BUS = 0
			if instrP1 == "Register":
				if registerArgument == 0:
					TEMP_BUS = Register0.Read()
				if registerArgument == 1:
					TEMP_BUS = Register1.Read()
				if registerArgument == 2:
					TEMP_BUS = Register2.Read()
				if registerArgument == 3:
					TEMP_BUS = Register3.Read()
			elif instrP1 == "MDR":
				TEMP_BUS = MDR.Read()
			elif instrP1 == "PC":
				TEMP_BUS = PC.Read()
			elif instrP1 == "Register F":
				TEMP_BUS = RegisterFlags.Read()
			elif instrP1 == "Address Register":
				TEMP_BUS = AddrR.Read()
			elif instrP1 == "Value Register":
				TEMP_BUS = ValuR.Read()
			elif instrP1 == "Accumulator":
				TEMP_BUS = RegisterY.Read()
			else:
				raise ValueError("Unknown Instruction 1 " + instrP1)

			if instrP2 == "Register":
				if registerArgument == 0:
					Register0.Write(TEMP_BUS)
				if registerArgument == 1:
					Register1.Write(TEMP_BUS)
				if registerArgument == 2:
					Register2.Write(TEMP_BUS)
				if registerArgument == 3:
					Register3.Write(TEMP_BUS)
			elif instrP2 == "MWR":
				MWR.Write(TEMP_BUS)
				RAM.Write(MAR.Read(), MWR.Read())
			elif instrP2 == "PC":
				PC.Write(TEMP_BUS)
			elif instrP2 == "Register F":
				RegisterFlags.Write(TEMP_BUS)
			elif instrP2 == "Address Register":
				AddrR.Write(TEMP_BUS)
			elif instrP2 == "Value Register":
				ValuR.Write(TEMP_BUS)
			elif instrP2 == "MAR":
				MAR.Write(TEMP_BUS)
				MDR.Write(preToDec(RAM.Read(MAR.Read())))
			elif instrP2 == "Register A":
				RegisterA.Write(TEMP_BUS)
			elif instrP2 == "Register B":
				RegisterB.Write(TEMP_BUS)
			else:
				raise ValueError("Unknown Instruction 2 " + instrP2)

			# stdscr.addstr(curses.LINES-2,0, "f" + MicroCodeDefinitions["MSB"][MicroInstruction[2]] + " t" + MicroCodeDefinitions["LSB"][MicroInstruction[3]])

		Micro_Counter.Inc()
		# stdscr.addstr(curses.LINES-2,0, MicroInstruction)
		return

def doCommand():
	global loops, prefix, clockPulses, InstructionsExecuted
	if HAVE_CURSES:
		stdscr.addstr(curses.LINES-1,0, prefix)
		cmd = stdscr.getstr(curses.LINES-1,1, 10).decode(encoding="utf-8")
		stdscr.addstr(curses.LINES-1,1, ' '*10)
		prefix = ':'
	else:
		cmd = input(":")

		
	if cmd.isdecimal():
		loops = int(cmd)
		return True
	elif cmd.lower() == "q":
		if HAVE_CURSES:
			curses.endwin()
		sys.exit()
		return True
	elif cmd == 'r':
		RAM.value = program.split()
		Main_Counter.Reset()
		Micro_Counter.Reset()
		PC.value = 255
		MAR.Reset()
		MDR.Reset()
		MWR.Reset()
		Register0.Reset()
		Register1.Reset()
		Register2.Reset()
		Register3.Reset()
		RegisterA.Reset()
		RegisterB.Reset()
		RegisterY.Reset()
		RegisterFlags.Reset()
		InstR.Reset()
		AddrR.Reset()
		ValuR.Reset()

		loops = 1
		clockPulses = 0
		InstructionsExecuted = 0
		return True
	elif cmd == "":
		return True
	else:
		if HAVE_CURSES:
			prefix = '?'
		else:
			print("CMD not known - ignored")
		return False

while not RegisterFlags.getFlag("H"):
	for i in range(loops):
		if not RegisterFlags.getFlag("H"):
			PrintDiagnostic()
			clock()
	loops = 1
	while not doCommand():
		pass

if HAVE_CURSES:
	PrintDiagnostic()
	prompt = "Done:"
	stdscr.addstr(curses.LINES-1,0, prompt)
	while stdscr.getstr(curses.LINES-1,len(prompt), 10).decode(encoding="utf-8").lower() not in ['ok', 'k', 'q']:
		pass
else:
	print("Done")